
#include "a4gl_libaubit4gl.h"
#include "a4gl_lib_lex_esqlc_int.h"
#include "field_handling.h"

#include "compile_c.h"
static int pr_when_do (char *when_str, int when_code, int l, char *f, char *when_to);

int when_code[8] = { WHEN_STOP,
  WHEN_NOTSET,
  WHEN_STOP,
  WHEN_CONTINUE,
  WHEN_CONTINUE,
  WHEN_CONTINUE,
  WHEN_NOTSET,
  WHEN_NOTSET
};
char when_to_tmp[64] = "";
char when_to[8][128] = { "", "", "", "", "", "", "", "" };


/**
 * Define the error handling after a certain point of the program.
 *
 * @param c The event to be catched:
 *  - WHEN_ERROR:
 *  - WHEN_ANYERROR:
 *  - WHEN_SQLERROR:
 *  - WHEN_WARNING:
 *  - WHEN_SQLWARNING:
 *  - WHEN_NOTFOUND:
 *  - WHEN_SUCCESS:
 *  - WHEN_SQLSUCCESS:
 * @param p The A4GL_action to execute.
 */
void
set_whenever (int c, char *p)
{
  int code;
  int oldcode;
  oldcode = c & 15;
  c = c >> 4;
  c = c << 4;
  code = -1;
  switch (c)
    {
    case WHEN_ERROR:
      set_whenever (WHEN_SQLERROR|oldcode, p);
      code = A_WHEN_ERROR;
      break;

    case WHEN_ANYERROR:
      set_whenever (WHEN_ERROR|oldcode, p);
      set_whenever (WHEN_SQLERROR|oldcode, p);
      code = A_WHEN_ANYERROR;
      break;

    case WHEN_SQLERROR:
      code = A_WHEN_SQLERROR;
      break;

    case WHEN_WARNING:
      code = A_WHEN_WARNING;
      break;

    case WHEN_SQLWARNING:
      code = A_WHEN_SQLWARNING;
      break;

    case WHEN_NOTFOUND:
      code = A_WHEN_NOTFOUND;
      break;


    case WHEN_SUCCESS:
      code = A_WHEN_SUCCESS;
      break;

    case WHEN_SQLSUCCESS:
      code = A_WHEN_SQLSUCCESS;
      break;
    }

  if (code == -1)
    {
      PRINTF ("Code=%d (%x) to %p\n", c, c, p);
      a4gl_yyerror ("Internal error setting whenever error...");
      exit (0);
    }

  if (p)
    {
      strcpy (when_to[code], p);
    }
  else
    {
      strcpy (when_to[code], when_to_tmp);
    }
  when_code[code] = oldcode;

  A4GL_whenchange(c);

  print_clr_status ();
}



/**
 * Called by the parser, in rules generated by the script mkyacc that joins all
 * the rules into one big yacc gramar.
 *
 * Generate the C code to output file that handles the error and warning
 * management controled by WHENEVER 4gl statememt.
 *
 * @param l The line number of the 4gl source code.
 */
void A4GL_prchkerr (int l,int isSql)
{
  //int a;
/* 0 = continue
 * 1 = stop
 * 2 = call
 * 3 = goto 
 */
  if (A4GL_isyes(acl_getenv("FUDGE_STATUS"))) {
  	printc("if (!aclfgli_get_err_flg()) {a4gl_status=0;}");
  }

  if (A4GL_doing_pcode () && 1)
    {
      char buff[2000];
      char tbuff[2000];
      SPRINTF1 (tbuff, "ERRCHK(%d,_module_name", l);
      strcpy (buff, tbuff);
      SPRINTF2 (tbuff, ",%d,\"%s\"", when_code[A_WHEN_SUCCESS], when_to[A_WHEN_SUCCESS]);
      strcat (buff, tbuff);
      SPRINTF2 (tbuff, ",%d,\"%s\"", when_code[A_WHEN_NOTFOUND], when_to[A_WHEN_NOTFOUND]);
      strcat (buff, tbuff);
      SPRINTF2 (tbuff, ",%d,\"%s\"", when_code[A_WHEN_SQLERROR], when_to[A_WHEN_SQLERROR]);
      strcat (buff, tbuff);
      SPRINTF2 (tbuff, ",%d,\"%s\"", when_code[A_WHEN_ERROR], when_to[A_WHEN_ERROR]);
      strcat (buff, tbuff);
      SPRINTF2 (tbuff, ",%d,\"%s\"", when_code[A_WHEN_WARNING], when_to[A_WHEN_WARNING]);
      strcat (buff, tbuff);
      SPRINTF0 (tbuff, ");");
      strcat (buff, tbuff);
      printc ("%s", buff);
      return;
    }

  printcomment ("/* NOTFOUND */");


  pr_when_do ("   ERR_CHK_WHEN_NOT_FOUND ", when_code[A_WHEN_NOTFOUND], l, 0, when_to[A_WHEN_NOTFOUND]);
  if (isSql) {
  	printcomment ("/* SQLERROR */");
  	pr_when_do ("   ERR_CHK_SQLERROR ", when_code[A_WHEN_SQLERROR], l, 0, when_to[A_WHEN_SQLERROR]);
  }

#ifdef ANYERRORISCAUSINGPROBS
  printcomment ("/* ANYERROR */");

    pr_when_do ("   if ( CHK_FOR_ERR && (a4gl_status<0||a4gl_sqlca.sqlcode<0))",
		when_code[A_WHEN_ANYERROR], l, f, when_to[A_WHEN_ANYERROR]);
#endif

  printcomment ("/* ERROR */");
  pr_when_do ("   ERR_CHK_ERROR ", when_code[A_WHEN_ERROR], l, 0, when_to[A_WHEN_ERROR]);

  if (isSql) {
  	printcomment ("/* SQLWARNING */");
  	pr_when_do ("   if (CHK_FOR_ERR && (a4gl_sqlca.sqlcode==0&&a4gl_sqlca.sqlawarn[0]=='W'))", when_code[A_WHEN_SQLWARNING], l, 0, when_to[A_WHEN_SQLWARNING]);
	}

  printcomment ("/* WARNING */");

  pr_when_do ("   ERR_CHK_WARNING ", when_code[A_WHEN_WARNING], l, 0, when_to[A_WHEN_WARNING]);



  	if ( when_code[A_WHEN_SUCCESS] == WHEN_CALL || when_code[A_WHEN_SQLSUCCESS] == WHEN_CALL) {
  		if (isSql) {
  				printcomment ("/* SQLSUCCESS */");
  				pr_when_do ("   if (a4gl_sqlca.sqlcode==0&&a4gl_status==0)", when_code[A_WHEN_SQLSUCCESS], l, 0, when_to[A_WHEN_SQLSUCCESS]);
  		}
  		printcomment ("/* SUCCESS */");
  		pr_when_do ("   if (a4gl_sqlca.sqlcode==0&&a4gl_status==0)", when_code[A_WHEN_SUCCESS], l, 0, when_to[A_WHEN_SUCCESS]);
  	}




}


/**
 * Print to the generated C code the specific instructions of WHENEVER 
 * statement implementation.
 *
 * @param when_str The string that contains the C comparison expresion
 *   to check if there was an error.
 * @param when_code The integer code of specific WHENEVER
 * @param l The line number on the 4gl source code.
 * @param f The 4gl source file name.
 * @param when_to
 */
static int pr_when_do (char *when_str, int when_code, int l, char *f, char *when_to)
{

  if ((when_code & 15) == WHEN_CONTINUE) {
		if (strstr(when_str,"ERR_CHK_ERROR")) {
      			printc ("%s { A4GL_err_continue_log(%d,_module_name); }\n", when_str, l);
		}
		if (strstr(when_str,"ERR_CHK_SQLERROR")) {
      			printc ("%s { A4GL_err_continue_log(%d,_module_name); }\n", when_str, l);
		}
    		return 0;
  }

  if ((when_code & 15) == WHEN_NOTSET)
    return 0;


  if (when_code == WHEN_STOP)
    {
	    
	  if (A4GL_doing_pcode()) {
      		printc ("%s A4GL_chk_err(%d,_module_name); \n", when_str, l);
	  } else {
      		printc ("%s { A4GL_chk_err(%d,_module_name); }\n", when_str, l);
      }
      printcomment ("/* WHENSTOP */");
    }

  if (when_code == WHEN_CALL)
    {
	char buff[256];
	strcpy(buff,when_to);	
	A4GL_convlower(buff);
      printc ("%s {A4GL_log_error(%d,_module_name,a4gl_status);%s%s(0); }\n", when_str, l,get_namespace (when_to), buff);
      add_function_to_header (when_to,get_namespace (when_to), 1,0);
      printcomment ("/* WHENCALL */");
    }

  if (when_code == WHEN_GOTO)
    {
	char buff[256];
	strcpy(buff,when_to);	
	A4GL_convlower(buff);
      printc ("%s {A4GL_log_error(%d,_module_name,a4gl_status); goto %s;}\n", when_str,l, buff);
      printcomment ("/* WHENGOTO */");
    }
  return 1;
}

